@@ -1,7 +1,8 @@
 import json
 import logging
 from asyncio import AbstractEventLoop
-from typing import Any, Text, List, Optional, Union, Dict
+from typing import Any, Text, List, Optional, Union
+import time
 
 from rasa.core.brokers.broker import EventBroker
 from rasa.shared.utils.io import DEFAULT_ENCODING
@@ -88,10 +89,38 @@ class KafkaEventBroker(EventBroker):
 
         return cls(broker_config.url, **broker_config.kwargs)
 
-    def publish(self, event) -> None:
-        self._create_producer()
-        self._publish(event)
-        self._close()
+    def publish(self, event, retries=60, retry_delay_in_seconds=5) -> None:
+        """Publishes events."""
+        if self.producer is None:
+            self._create_producer()
+            connected = self.producer.bootstrap_connected()
+            if connected:
+                logger.debug("Connection to kafka successful.")
+            else:
+                logger.debug("Failed to connect kafka.")
+                return
+        while retries:
+            try:
+                self._publish(event)
+                return
+            except Exception as e:
+                logger.error(
+                    f"Could not publish message to kafka host '{self.host}'. "
+                    f"Failed with error: {e}"
+                )
+                connected = self.producer.bootstrap_connected()
+                if not connected:
+                    self._close()
+                    logger.debug("Connection to kafka lost, reconnecting...")
+                    self._create_producer()
+                    connected = self.producer.bootstrap_connected()
+                    if connected:
+                        logger.debug("Reconnection to kafka successful")
+                        self._publish(event)
+                retries -= 1
+                time.sleep(retry_delay_in_seconds)
+
+        logger.error("Failed to publish Kafka event.")
 
     def _create_producer(self) -> None:
         import kafka
@@ -146,6 +175,7 @@ class KafkaEventBroker(EventBroker):
             )
 
     def _publish(self, event) -> None:
+        logger.debug(f"Calling kafka send({self.topic}, {event})")
         self.producer.send(self.topic, event)
 
     def _close(self) -> None:
