@@ -22,6 +22,8 @@ from rasa.shared.constants import (
     DOCS_URL_POLICIES,
     DOCS_URL_MIGRATION_GUIDE,
     DEFAULT_CONFIG_PATH,
+    DOCS_URL_ACTIONS,
+    DOCS_URL_DEFAULT_ACTIONS,
 )
 from rasa.shared.core.constants import (
     USER_INTENT_BACK,
@@ -73,10 +75,13 @@ class PolicyEnsemble:
         if not any(
             isinstance(policy, (MappingPolicy, RulePolicy)) for policy in self.policies
         ):
-            logger.info(
-                f"MappingPolicy not included in policy ensemble. Default intents "
-                f"'{USER_INTENT_RESTART} and {USER_INTENT_BACK} will not trigger "
-                f"actions '{ACTION_RESTART_NAME}' and '{ACTION_BACK_NAME}'."
+            rasa.shared.utils.io.raise_warning(
+                f"Neither '{RulePolicy.__name__}' nor '{MappingPolicy.__name__}' "
+                f"(deprecated) are included in the model's policy configuration. "
+                f"Default intents such as '{USER_INTENT_RESTART}' and "
+                f"'{USER_INTENT_BACK}' will not trigger actions "
+                f"'{ACTION_RESTART_NAME}' and '{ACTION_BACK_NAME}'.",
+                docs=DOCS_URL_DEFAULT_ACTIONS,
             )
 
     @staticmethod
