@@ -15,6 +15,7 @@ from typing import (
     Union,
     Dict,
     Any,
+    NamedTuple,
 )
 
 from rasa import telemetry
@@ -64,7 +65,14 @@ ENTITY_PROCESSORS = {"EntitySynonymMapper", "ResponseSelector"}
 
 EXTRACTORS_WITH_CONFIDENCES = {"CRFEntityExtractor", "DIETClassifier"}
 
-CVEvaluationResult = namedtuple("Results", "train test")
+
+class CVEvaluationResult(NamedTuple):
+    """Stores NLU cross-validation results."""
+
+    train: Dict
+    test: Dict
+    evaluation: Dict
+
 
 NO_ENTITY = "no_entity"
 
@@ -201,16 +209,27 @@ def write_intent_successes(
         logger.info("No successful intent predictions found.")
 
 
-def write_intent_errors(
-    intent_results: List[IntentEvaluationResult], errors_filename: Text
-) -> None:
+def _write_errors(errors: List[Dict], errors_filename: Text, error_type: Text) -> None:
     """Write incorrect intent predictions to a file.
 
     Args:
-        intent_results: intent evaluation result
+        errors: Serializable prediction errors.
         errors_filename: filename of file to save incorrect predictions to
+        error_type: NLU entity which was evaluated (e.g. `intent` or `entity`).
     """
-    errors = [
+    if errors:
+        rasa.shared.utils.io.dump_obj_as_json_to_file(errors_filename, errors)
+        logger.info(f"Incorrect {error_type} predictions saved to {errors_filename}.")
+        logger.debug(
+            f"\n\nThese {error_type} examples could not be classified "
+            f"correctly: \n{errors}"
+        )
+    else:
+        logger.info(f"Every {error_type} was predicted correctly by the model.")
+
+
+def _get_intent_errors(intent_results: List[IntentEvaluationResult]) -> List[Dict]:
+    return [
         {
             "text": r.message,
             "intent": r.intent_target,
@@ -223,16 +242,6 @@ def write_intent_errors(
         if r.intent_target != r.intent_prediction
     ]
 
-    if errors:
-        rasa.shared.utils.io.dump_obj_as_json_to_file(errors_filename, errors)
-        logger.info(f"Incorrect intent predictions saved to {errors_filename}.")
-        logger.debug(
-            "\n\nThese intent examples could not be classified "
-            "correctly: \n{}".format(errors)
-        )
-    else:
-        logger.info("Your model predicted all intents successfully.")
-
 
 def write_response_successes(
     response_results: List[ResponseSelectionEvaluationResult], successes_filename: Text
@@ -267,16 +276,18 @@ def write_response_successes(
         logger.info("No successful response predictions found.")
 
 
-def write_response_errors(
-    response_results: List[ResponseSelectionEvaluationResult], errors_filename: Text
-) -> None:
+def _response_errors(
+    response_results: List[ResponseSelectionEvaluationResult],
+) -> List[Dict]:
     """Write incorrect response selection predictions to a file.
 
     Args:
         response_results: response selection evaluation result
-        errors_filename: filename of file to save incorrect predictions to
+
+    Returns:
+        Serializable prediction errors.
     """
-    errors = [
+    return [
         {
             "text": r.message,
             "intent_response_key_target": r.intent_response_key_target,
@@ -289,16 +300,6 @@ def write_response_errors(
         if r.intent_response_key_prediction != r.intent_response_key_target
     ]
 
-    if errors:
-        rasa.shared.utils.io.dump_obj_as_json_to_file(errors_filename, errors)
-        logger.info(f"Incorrect response predictions saved to {errors_filename}.")
-        logger.debug(
-            "\n\nThese response examples could not be classified "
-            "correctly: \n{}".format(errors)
-        )
-    else:
-        logger.info("Your model predicted all responses successfully.")
-
 
 def plot_attribute_confidences(
     results: Union[
@@ -340,13 +341,13 @@ def plot_entity_confidences(
     hist_filename: Text,
     title: Text,
 ) -> None:
-    """Create histogram of confidence distribution.
+    """Creates histogram of confidence distribution.
 
     Args:
-        results: evaluation results
+        merged_targets: Entity labels.
+        merged_predictions: Predicted entities.
+        merged_confidences: Confidence scores of predictions.
         hist_filename: filename to save plot to
-        target_key: key of target in results
-        prediction_key: key of predictions in results
         title: title of plot
     """
     pos_hist = [
@@ -374,6 +375,7 @@ def evaluate_response_selections(
     successes: bool,
     errors: bool,
     disable_plotting: bool,
+    report_as_dict: Optional[bool] = None,
 ) -> Dict:  # pragma: no cover
     """Creates summary statistics for response selection.
 
@@ -387,13 +389,13 @@ def evaluate_response_selections(
         successes: if True success are written down to disk
         errors: if True errors are written down to disk
         disable_plotting: if True no plots are created
+        report_as_dict: `True` if the evaluation report should be returned as `dict`.
+            If `False` the report is returned in a human-readable text format. If `None`
+            `report_as_dict` is considered as `True` in case an `output_directory` is
+            given.
 
     Returns: dictionary with evaluation results
     """
-    import sklearn.metrics
-    import sklearn.utils.multiclass
-    from rasa.test import get_evaluation_metrics
-
     # remove empty response targets
     num_examples = len(response_selection_results)
     response_selection_results = remove_empty_response_examples(
@@ -415,33 +417,14 @@ def evaluate_response_selections(
         "intent_response_key_prediction",
     )
 
-    confusion_matrix = sklearn.metrics.confusion_matrix(
-        target_intent_response_keys, predicted_intent_response_keys
+    report, precision, f1, accuracy, confusion_matrix, labels = _calculate_report(
+        output_directory,
+        target_intent_response_keys,
+        predicted_intent_response_keys,
+        report_as_dict,
     )
-    labels = sklearn.utils.multiclass.unique_labels(
-        target_intent_response_keys, predicted_intent_response_keys
-    )
-
     if output_directory:
-        report, precision, f1, accuracy = get_evaluation_metrics(
-            target_intent_response_keys,
-            predicted_intent_response_keys,
-            output_dict=True,
-        )
-        report = _add_confused_labels_to_report(report, confusion_matrix, labels)
-
-        report_filename = os.path.join(
-            output_directory, "response_selection_report.json"
-        )
-        rasa.shared.utils.io.dump_obj_as_json_to_file(report_filename, report)
-        logger.info(f"Classification report saved to {report_filename}.")
-
-    else:
-        report, precision, f1, accuracy = get_evaluation_metrics(
-            target_intent_response_keys, predicted_intent_response_keys
-        )
-        if isinstance(report, str):
-            log_evaluation_table(report, precision, f1, accuracy)
+        _dump_report(output_directory, "response_selection_report.json", report)
 
     if successes:
         successes_filename = "response_selection_successes.json"
@@ -450,12 +433,12 @@ def evaluate_response_selections(
         # save classified samples to file for debugging
         write_response_successes(response_selection_results, successes_filename)
 
-    if errors:
+    response_errors = _response_errors(response_selection_results)
+
+    if errors and output_directory:
         errors_filename = "response_selection_errors.json"
-        if output_directory:
-            errors_filename = os.path.join(output_directory, errors_filename)
-        # log and save misclassified samples to file for debugging
-        write_response_errors(response_selection_results, errors_filename)
+        errors_filename = os.path.join(output_directory, errors_filename)
+        _write_errors(response_errors, errors_filename, error_type="response")
 
     if not disable_plotting:
         confusion_matrix_filename = "response_selection_confusion_matrix.png"
@@ -498,6 +481,7 @@ def evaluate_response_selections(
         "precision": precision,
         "f1_score": f1,
         "accuracy": accuracy,
+        "errors": response_errors,
     }
 
 
@@ -560,6 +544,7 @@ def evaluate_intents(
     successes: bool,
     errors: bool,
     disable_plotting: bool,
+    report_as_dict: Optional[bool] = None,
 ) -> Dict:  # pragma: no cover
     """Creates summary statistics for intents.
 
@@ -572,13 +557,13 @@ def evaluate_intents(
         successes: if True correct predictions are written to disk
         errors: if True incorrect predictions are written to disk
         disable_plotting: if True no plots are created
+        report_as_dict: `True` if the evaluation report should be returned as `dict`.
+            If `False` the report is returned in a human-readable text format. If `None`
+            `report_as_dict` is considered as `True` in case an `output_directory` is
+            given.
 
     Returns: dictionary with evaluation results
     """
-    import sklearn.metrics
-    import sklearn.utils.multiclass
-    from rasa.test import get_evaluation_metrics
-
     # remove empty intent targets
     num_examples = len(intent_results)
     intent_results = remove_empty_intent_examples(intent_results)
@@ -592,37 +577,21 @@ def evaluate_intents(
         intent_results, "intent_target", "intent_prediction"
     )
 
-    confusion_matrix = sklearn.metrics.confusion_matrix(
-        target_intents, predicted_intents
+    report, precision, f1, accuracy, confusion_matrix, labels = _calculate_report(
+        output_directory, target_intents, predicted_intents, report_as_dict,
     )
-    labels = sklearn.utils.multiclass.unique_labels(target_intents, predicted_intents)
-
     if output_directory:
-        report, precision, f1, accuracy = get_evaluation_metrics(
-            target_intents, predicted_intents, output_dict=True
-        )
-        report = _add_confused_labels_to_report(report, confusion_matrix, labels)
-
-        report_filename = os.path.join(output_directory, "intent_report.json")
-        rasa.shared.utils.io.dump_obj_as_json_to_file(report_filename, report)
-        logger.info(f"Classification report saved to {report_filename}.")
-
-    else:
-        report, precision, f1, accuracy = get_evaluation_metrics(
-            target_intents, predicted_intents
-        )
-        if isinstance(report, str):
-            log_evaluation_table(report, precision, f1, accuracy)
+        _dump_report(output_directory, "intent_report.json", report)
 
     if successes and output_directory:
         successes_filename = os.path.join(output_directory, "intent_successes.json")
         # save classified samples to file for debugging
         write_intent_successes(intent_results, successes_filename)
 
+    intent_errors = _get_intent_errors(intent_results)
     if errors and output_directory:
         errors_filename = os.path.join(output_directory, "intent_errors.json")
-        # log and save misclassified samples to file for debugging
-        write_intent_errors(intent_results, errors_filename)
+        _write_errors(intent_errors, errors_filename, "intent")
 
     if not disable_plotting:
         confusion_matrix_filename = "intent_confusion_matrix.png"
@@ -664,9 +633,59 @@ def evaluate_intents(
         "precision": precision,
         "f1_score": f1,
         "accuracy": accuracy,
+        "errors": intent_errors,
     }
 
 
+def _calculate_report(
+    output_directory: Optional[Text],
+    targets: Iterable[Any],
+    predictions: Iterable[Any],
+    report_as_dict: Optional[bool] = None,
+    exclude_label: Optional[Text] = None,
+) -> Tuple[Union[Text, Dict], float, float, float, np.ndarray, List[Text]]:
+    from rasa.test import get_evaluation_metrics
+    import sklearn.metrics
+    import sklearn.utils.multiclass
+
+    confusion_matrix = sklearn.metrics.confusion_matrix(targets, predictions)
+    labels = sklearn.utils.multiclass.unique_labels(targets, predictions)
+
+    if report_as_dict is None:
+        report_as_dict = bool(output_directory)
+
+    if output_directory:
+        report, precision, f1, accuracy = get_evaluation_metrics(
+            targets,
+            predictions,
+            output_dict=report_as_dict,
+            exclude_label=exclude_label,
+        )
+        report = _add_confused_labels_to_report(
+            report,
+            confusion_matrix,
+            labels,
+            exclude_labels=[exclude_label] if exclude_label else [],
+        )
+    else:
+        report, precision, f1, accuracy = get_evaluation_metrics(
+            targets,
+            predictions,
+            output_dict=report_as_dict,
+            exclude_label=exclude_label,
+        )
+        if isinstance(report, str):
+            log_evaluation_table(report, precision, f1, accuracy)
+
+    return report, precision, f1, accuracy, confusion_matrix, labels
+
+
+def _dump_report(output_directory: Text, filename: Text, report: Dict) -> None:
+    report_filename = os.path.join(output_directory, filename)
+    rasa.shared.utils.io.dump_obj_as_json_to_file(report_filename, report)
+    logger.info(f"Classification report saved to {report_filename}.")
+
+
 def merge_labels(
     aligned_predictions: List[Dict], extractor: Optional[Text] = None
 ) -> List[Text]:
@@ -722,35 +741,6 @@ def substitute_labels(labels: List[Text], old: Text, new: Text) -> List[Text]:
     return [new if label == old else label for label in labels]
 
 
-def write_incorrect_entity_predictions(
-    entity_results: List[EntityEvaluationResult],
-    merged_targets: List[Text],
-    merged_predictions: List[Text],
-    error_filename: Text,
-) -> None:
-    """Write incorrect entity predictions to a file.
-
-    Args:
-        entity_results: response selection evaluation result
-        merged_predictions: list of predicted entity labels
-        merged_targets: list of true entity labels
-        error_filename: filename of file to save incorrect predictions to
-    """
-    errors = collect_incorrect_entity_predictions(
-        entity_results, merged_predictions, merged_targets
-    )
-
-    if errors:
-        rasa.shared.utils.io.dump_obj_as_json_to_file(error_filename, errors)
-        logger.info(f"Incorrect entity predictions saved to {error_filename}.")
-        logger.debug(
-            "\n\nThese intent examples could not be classified "
-            "correctly: \n{}".format(errors)
-        )
-    else:
-        logger.info("Your model predicted all entities successfully.")
-
-
 def collect_incorrect_entity_predictions(
     entity_results: List[EntityEvaluationResult],
     merged_predictions: List[Text],
@@ -851,6 +841,7 @@ def evaluate_entities(
     successes: bool,
     errors: bool,
     disable_plotting: bool,
+    report_as_dict: Optional[bool] = None,
 ) -> Dict:  # pragma: no cover
     """Creates summary statistics for each entity extractor.
 
@@ -863,13 +854,13 @@ def evaluate_entities(
         successes: if True correct predictions are written to disk
         errors: if True incorrect predictions are written to disk
         disable_plotting: if True no plots are created
+        report_as_dict: `True` if the evaluation report should be returned as `dict`.
+            If `False` the report is returned in a human-readable text format. If `None`
+            `report_as_dict` is considered as `True` in case an `output_directory` is
+            given.
 
     Returns: dictionary with evaluation results
     """
-    import sklearn.metrics
-    import sklearn.utils.multiclass
-    from rasa.test import get_evaluation_metrics
-
     aligned_predictions = align_all_entity_predictions(entity_results, extractors)
     merged_targets = merge_labels(aligned_predictions)
     merged_targets = substitute_labels(merged_targets, NO_ENTITY_TAG, NO_ENTITY)
@@ -884,45 +875,15 @@ def evaluate_entities(
 
         logger.info(f"Evaluation for entity extractor: {extractor} ")
 
-        confusion_matrix = sklearn.metrics.confusion_matrix(
-            merged_targets, merged_predictions
+        report, precision, f1, accuracy, confusion_matrix, labels = _calculate_report(
+            output_directory,
+            merged_targets,
+            merged_predictions,
+            report_as_dict,
+            exclude_label=NO_ENTITY,
         )
-        labels = sklearn.utils.multiclass.unique_labels(
-            merged_targets, merged_predictions
-        )
-
         if output_directory:
-            report_filename = f"{extractor}_report.json"
-            extractor_report_filename = os.path.join(output_directory, report_filename)
-
-            report, precision, f1, accuracy = get_evaluation_metrics(
-                merged_targets,
-                merged_predictions,
-                output_dict=True,
-                exclude_label=NO_ENTITY,
-            )
-            report = _add_confused_labels_to_report(
-                report, confusion_matrix, labels, [NO_ENTITY]
-            )
-
-            rasa.shared.utils.io.dump_obj_as_json_to_file(
-                extractor_report_filename, report
-            )
-
-            logger.info(
-                "Classification report for '{}' saved to '{}'."
-                "".format(extractor, extractor_report_filename)
-            )
-
-        else:
-            report, precision, f1, accuracy = get_evaluation_metrics(
-                merged_targets,
-                merged_predictions,
-                output_dict=False,
-                exclude_label=NO_ENTITY,
-            )
-            if isinstance(report, str):
-                log_evaluation_table(report, precision, f1, accuracy)
+            _dump_report(output_directory, f"{extractor}_report.json", report)
 
         if successes:
             successes_filename = f"{extractor}_successes.json"
@@ -933,14 +894,13 @@ def evaluate_entities(
                 entity_results, merged_targets, merged_predictions, successes_filename
             )
 
-        if errors:
-            errors_filename = f"{extractor}_errors.json"
-            if output_directory:
-                errors_filename = os.path.join(output_directory, errors_filename)
-            # log and save misclassified samples to file for debugging
-            write_incorrect_entity_predictions(
-                entity_results, merged_targets, merged_predictions, errors_filename
-            )
+        entity_errors = collect_incorrect_entity_predictions(
+            entity_results, merged_predictions, merged_targets
+        )
+        if errors and output_directory:
+            errors_filename = os.path.join(output_directory, f"{extractor}_errors.json")
+
+            _write_errors(entity_errors, errors_filename, "entity")
 
         if not disable_plotting:
             confusion_matrix_filename = f"{extractor}_confusion_matrix.png"
@@ -975,6 +935,7 @@ def evaluate_entities(
             "precision": precision,
             "f1_score": f1,
             "accuracy": accuracy,
+            "errors": entity_errors,
         }
 
     return result
@@ -1462,6 +1423,7 @@ def run_evaluation(
     errors: bool = False,
     component_builder: Optional[ComponentBuilder] = None,
     disable_plotting: bool = False,
+    report_as_dict: Optional[bool] = None,
 ) -> Dict:  # pragma: no cover
     """Evaluate intent classification, response selection and entity extraction.
 
@@ -1473,6 +1435,10 @@ def run_evaluation(
         errors: if true incorrect predictions are written to a file
         component_builder: component builder
         disable_plotting: if true confusion matrix and histogram will not be rendered
+        report_as_dict: `True` if the evaluation report should be returned as `dict`.
+            If `False` the report is returned in a human-readable text format. If `None`
+            `report_as_dict` is considered as `True` in case an `output_directory` is
+            given.
 
     Returns: dictionary containing evaluation results
     """
@@ -1502,7 +1468,12 @@ def run_evaluation(
     if intent_results:
         logger.info("Intent evaluation results:")
         result["intent_evaluation"] = evaluate_intents(
-            intent_results, output_directory, successes, errors, disable_plotting
+            intent_results,
+            output_directory,
+            successes,
+            errors,
+            disable_plotting,
+            report_as_dict=report_as_dict,
         )
 
     if response_selection_results:
@@ -1513,6 +1484,7 @@ def run_evaluation(
             successes,
             errors,
             disable_plotting,
+            report_as_dict=report_as_dict,
         )
 
     if any(entity_results):
@@ -1525,6 +1497,7 @@ def run_evaluation(
             successes,
             errors,
             disable_plotting,
+            report_as_dict=report_as_dict,
         )
 
     telemetry.track_nlu_model_test(test_data)
@@ -1638,11 +1611,12 @@ def _contains_entity_labels(entity_results: List[EntityEvaluationResult]) -> boo
 def cross_validate(
     data: TrainingData,
     n_folds: int,
-    nlu_config: Union[RasaNLUModelConfig, Text],
+    nlu_config: Union[RasaNLUModelConfig, Text, Dict],
     output: Optional[Text] = None,
     successes: bool = False,
     errors: bool = False,
     disable_plotting: bool = False,
+    report_as_dict: Optional[bool] = None,
 ) -> Tuple[CVEvaluationResult, CVEvaluationResult, CVEvaluationResult]:
     """Stratified cross validation on data.
 
@@ -1654,6 +1628,10 @@ def cross_validate(
         successes: if true successful predictions are written to a file
         errors: if true incorrect predictions are written to a file
         disable_plotting: if true no confusion matrix and historgram plates are created
+        report_as_dict: `True` if the evaluation report should be returned as `dict`.
+            If `False` the report is returned in a human-readable text format. If `None`
+            `report_as_dict` is considered as `True` in case an `output_directory` is
+            given.
 
     Returns:
         dictionary with key, list structure, where each entry in list
@@ -1661,7 +1639,7 @@ def cross_validate(
     """
     import rasa.nlu.config
 
-    if isinstance(nlu_config, str):
+    if isinstance(nlu_config, (str, Dict)):
         nlu_config = rasa.nlu.config.load(nlu_config)
 
     if output:
@@ -1721,22 +1699,41 @@ def cross_validate(
         if is_response_selector_present(interpreter):
             response_selector_present = True
 
+    intent_evaluation = {}
     if intent_classifier_present and intent_test_results:
         logger.info("Accumulated test folds intent evaluation results:")
-        evaluate_intents(
-            intent_test_results, output, successes, errors, disable_plotting
+        intent_evaluation = evaluate_intents(
+            intent_test_results,
+            output,
+            successes,
+            errors,
+            disable_plotting,
+            report_as_dict=report_as_dict,
         )
 
+    entity_evaluation = {}
     if extractors and entity_evaluation_possible:
         logger.info("Accumulated test folds entity evaluation results:")
-        evaluate_entities(
-            entity_test_results, extractors, output, successes, errors, disable_plotting
+        entity_evaluation = evaluate_entities(
+            entity_test_results,
+            extractors,
+            output,
+            successes,
+            errors,
+            disable_plotting,
+            report_as_dict=report_as_dict,
         )
 
+    responses_evaluation = {}
     if response_selector_present and response_selection_test_results:
         logger.info("Accumulated test folds response selection evaluation results:")
-        evaluate_response_selections(
-            response_selection_test_results, output, successes, errors, disable_plotting
+        responses_evaluation = evaluate_response_selections(
+            response_selection_test_results,
+            output,
+            successes,
+            errors,
+            disable_plotting,
+            report_as_dict=report_as_dict,
         )
 
     if not entity_evaluation_possible:
@@ -1744,11 +1741,16 @@ def cross_validate(
         entity_train_metrics = defaultdict(lambda: defaultdict(list))
 
     return (
-        CVEvaluationResult(dict(intent_train_metrics), dict(intent_test_metrics)),
-        CVEvaluationResult(dict(entity_train_metrics), dict(entity_test_metrics)),
+        CVEvaluationResult(
+            dict(intent_train_metrics), dict(intent_test_metrics), intent_evaluation
+        ),
+        CVEvaluationResult(
+            dict(entity_train_metrics), dict(entity_test_metrics), entity_evaluation
+        ),
         CVEvaluationResult(
             dict(response_selection_train_metrics),
             dict(response_selection_test_metrics),
+            responses_evaluation,
         ),
     )
 
